@@ -66,8 +66,6 @@ class Gem::Installer
 
   attr_reader :package
 
-  @path_warning = false
-
   class << self
     #
     # Changes in rubygems to lazily loading `rubygems/command` (in order to
@@ -86,11 +84,6 @@ class Gem::Installer
       super(klass)
     end
 
-    ##
-    # True if we've warned about PATH not including Gem.bindir
-
-    attr_accessor :path_warning
-
     ##
     # Overrides the executable format.
     #
@@ -188,13 +181,6 @@ class Gem::Installer
     @package.dir_mode = options[:dir_mode]
     @package.prog_mode = options[:prog_mode]
     @package.data_mode = options[:data_mode]
-
-    if options[:user_install]
-      @gem_home = Gem.user_dir
-      @bin_dir = Gem.bindir gem_home unless options[:bin_dir]
-      @plugins_dir = Gem.plugindir(gem_home)
-      check_that_user_bin_dir_is_in_path
-    end
   end
 
   ##
@@ -222,10 +208,11 @@ class Gem::Installer
 
     File.open generated_bin, "rb" do |io|
       line = io.gets
-      shebang = /^#!.*ruby/
+      shebang = /^#!.*ruby/o
 
-      if load_relative_enabled?
-        until line.nil? || line =~ shebang do
+      # TruffleRuby uses a bash prelude in default launchers
+      if load_relative_enabled? || RUBY_ENGINE == "truffleruby"
+        until line.nil? || shebang.match?(line) do
           line = io.gets
         end
       end
@@ -236,17 +223,17 @@ class Gem::Installer
 
       # TODO: detect a specially formatted comment instead of trying
       # to find a string inside Ruby code.
-      next unless io.gets.to_s.include?("This file was generated by RubyGems")
+      next unless io.gets&.include?("This file was generated by RubyGems")
 
       ruby_executable = true
-      existing = io.read.slice(%r{
+      existing = io.read.slice(/
           ^\s*(
             gem \s |
             load \s Gem\.bin_path\( |
             load \s Gem\.activate_bin_path\(
           )
           (['"])(.*?)(\2),
-        }x, 3)
+        /x, 3)
     end
 
     return if spec.name == existing
@@ -316,7 +303,7 @@ class Gem::Installer
     FileUtils.rm_rf spec.extension_dir
 
     dir_mode = options[:dir_mode]
-    FileUtils.mkdir_p gem_dir, :mode => dir_mode && 0o755
+    FileUtils.mkdir_p gem_dir, mode: dir_mode && 0o755
 
     if @options[:install_as_default]
       extract_bin
@@ -341,13 +328,16 @@ class Gem::Installer
 
     say spec.post_install_message if options[:post_install_message] && !spec.post_install_message.nil?
 
-    Gem::Specification.add_spec(spec)
+    Gem::Specification.add_spec(spec) unless @install_dir
 
     load_plugin
 
     run_post_install_hooks
 
     spec
+  rescue Errno::EACCES => e
+    # Permission denied - /path/to/foo
+    raise Gem::FilePermissionError, e.message.split(" - ").last
   end
 
   def run_pre_install_hooks # :nodoc:
@@ -388,7 +378,7 @@ class Gem::Installer
       specs = []
 
       Gem::Util.glob_files_in_dir("*.gemspec", File.join(gem_home, "specifications")).each do |path|
-        spec = Gem::Specification.load path.tap(&Gem::UNTAINT)
+        spec = Gem::Specification.load path
         specs << spec if spec
       end
 
@@ -458,6 +448,9 @@ class Gem::Installer
   ##
   # Writes the full .gemspec specification (in Ruby) to the gem home's
   # specifications/default directory.
+  #
+  # In contrast to #write_spec, this keeps file lists, so the `gem contents`
+  # command works.
 
   def write_default_spec
     Gem.write_binary(default_spec_file, spec.to_ruby)
@@ -479,12 +472,21 @@ class Gem::Installer
   end
 
   def generate_bin # :nodoc:
-    return if spec.executables.nil? || spec.executables.empty?
+    executables = spec.executables
+    return if executables.nil? || executables.empty?
+
+    if @gem_home == Gem.user_dir
+      # If we get here, then one of the following likely happened:
+      # - `--user-install` was specified
+      # - `Gem::PathSupport#home` fell back to `Gem.user_dir`
+      # - GEM_HOME was manually set to `Gem.user_dir`
+
+      check_that_user_bin_dir_is_in_path(executables)
+    end
 
     ensure_writable_dir @bin_dir
 
-    spec.executables.each do |filename|
-      filename.tap(&Gem::UNTAINT)
+    executables.each do |filename|
       bin_path = File.join gem_dir, spec.bindir, filename
       next unless File.exist? bin_path
 
@@ -492,8 +494,7 @@ class Gem::Installer
       dir_mode = options[:prog_mode] || (mode | 0o111)
 
       unless dir_mode == mode
-        require "fileutils"
-        FileUtils.chmod dir_mode, bin_path
+        File.chmod dir_mode, bin_path
       end
 
       check_executable_overwrite filename
@@ -517,6 +518,8 @@ class Gem::Installer
     else
       regenerate_plugins_for(spec, @plugins_dir)
     end
+  rescue ArgumentError => e
+    raise e, "#{latest.name} #{latest.version} #{spec.name} #{spec.version}: #{e.message}"
   end
 
   ##
@@ -524,17 +527,19 @@ class Gem::Installer
   #--
   # The Windows script is generated in addition to the regular one due to a
   # bug or misfeature in the Windows shell's pipe.  See
-  # http://blade.nagaokaut.ac.jp/cgi-bin/scat.rb/ruby/ruby-talk/193379
+  # https://blade.ruby-lang.org/ruby-talk/193379
 
   def generate_bin_script(filename, bindir)
     bin_script_path = File.join bindir, formatted_program_filename(filename)
 
-    require "fileutils"
-    FileUtils.rm_f bin_script_path # prior install may have been --no-wrappers
+    Gem.open_file_with_lock(bin_script_path) do
+      require "fileutils"
+      FileUtils.rm_f bin_script_path # prior install may have been --no-wrappers
 
-    File.open bin_script_path, "wb", 0o755 do |file|
-      file.print app_script_text(filename)
-      file.chmod(options[:prog_mode] || 0o755)
+      File.open(bin_script_path, "wb", 0o755) do |file|
+        file.write app_script_text(filename)
+        file.chmod(options[:prog_mode] || 0o755)
+      end
     end
 
     verbose bin_script_path
@@ -559,7 +564,7 @@ class Gem::Installer
       File.unlink dst
     end
 
-    FileUtils.symlink src, dst, :verbose => Gem.configuration.really_verbose
+    FileUtils.symlink src, dst, verbose: Gem.configuration.really_verbose
   rescue NotImplementedError, SystemCallError
     alert_warning "Unable to use symlinks, installing wrapper"
     generate_bin_script filename, bindir
@@ -582,7 +587,7 @@ class Gem::Installer
 
   def shebang(bin_file_name)
     path = File.join gem_dir, spec.bindir, bin_file_name
-    first_line = File.open(path, "rb") {|file| file.gets } || ""
+    first_line = File.open(path, "rb", &:gets) || ""
 
     if first_line.start_with?("#!")
       # Preserve extra words on shebang line, like "-w".  Thanks RPA.
@@ -624,7 +629,6 @@ class Gem::Installer
 
   def ensure_loadable_spec
     ruby = spec.to_ruby_for_cache
-    ruby.tap(&Gem::UNTAINT)
 
     begin
       eval ruby
@@ -645,32 +649,37 @@ class Gem::Installer
 
   def process_options # :nodoc:
     @options = {
-      :bin_dir => nil,
-      :env_shebang => false,
-      :force => false,
-      :only_install_dir => false,
-      :post_install_message => true,
+      bin_dir: nil,
+      env_shebang: false,
+      force: false,
+      only_install_dir: false,
+      post_install_message: true,
     }.merge options
 
     @env_shebang         = options[:env_shebang]
     @force               = options[:force]
     @install_dir         = options[:install_dir]
-    @gem_home            = options[:install_dir] || Gem.dir
-    @plugins_dir         = Gem.plugindir(@gem_home)
+    @user_install        = options[:user_install]
     @ignore_dependencies = options[:ignore_dependencies]
     @format_executable   = options[:format_executable]
     @wrappers            = options[:wrappers]
     @only_install_dir    = options[:only_install_dir]
 
-    # If the user has asked for the gem to be installed in a directory that is
-    # the system gem directory, then use the system bin directory, else create
-    # (or use) a new bin dir under the gem_home.
-    @bin_dir             = options[:bin_dir] || Gem.bindir(gem_home)
+    @bin_dir             = options[:bin_dir]
     @development         = options[:development]
     @build_root          = options[:build_root]
 
     @build_args = options[:build_args]
 
+    @gem_home = @install_dir || user_install_dir || Gem.dir
+
+    # If the user has asked for the gem to be installed in a directory that is
+    # the system gem directory, then use the system bin directory, else create
+    # (or use) a new bin dir under the gem_home.
+    @bin_dir ||= Gem.bindir(@gem_home)
+
+    @plugins_dir = Gem.plugindir(@gem_home)
+
     unless @build_root.nil?
       @bin_dir = File.join(@build_root, @bin_dir.gsub(/^[a-zA-Z]:/, ""))
       @gem_home = File.join(@build_root, @gem_home.gsub(/^[a-zA-Z]:/, ""))
@@ -679,9 +688,7 @@ class Gem::Installer
     end
   end
 
-  def check_that_user_bin_dir_is_in_path # :nodoc:
-    return if self.class.path_warning
-
+  def check_that_user_bin_dir_is_in_path(executables) # :nodoc:
     user_bin_dir = @bin_dir || Gem.bindir(gem_home)
     user_bin_dir = user_bin_dir.tr(File::ALT_SEPARATOR, File::SEPARATOR) if File::ALT_SEPARATOR
 
@@ -697,15 +704,13 @@ class Gem::Installer
 
     unless path.include? user_bin_dir
       unless !Gem.win_platform? && (path.include? user_bin_dir.sub(ENV["HOME"], "~"))
-        alert_warning "You don't have #{user_bin_dir} in your PATH,\n\t  gem executables will not run."
-        self.class.path_warning = true
+        alert_warning "You don't have #{user_bin_dir} in your PATH,\n\t  gem executables (#{executables.join(", ")}) will not run."
       end
     end
   end
 
   def verify_gem_home # :nodoc:
-    FileUtils.mkdir_p gem_home, :mode => options[:dir_mode] && 0o755
-    raise Gem::FilePermissionError, gem_home unless File.writable?(gem_home)
+    FileUtils.mkdir_p gem_home, mode: options[:dir_mode] && 0o755
   end
 
   def verify_spec
@@ -804,7 +809,7 @@ TEXT
     ruby_exe = "#{rb_config["RUBY_INSTALL_NAME"]}#{rb_config["EXEEXT"]}"
     ruby_exe = "ruby.exe" if ruby_exe.empty?
 
-    if File.exist?(File.join bindir, ruby_exe)
+    if File.exist?(File.join(bindir, ruby_exe))
       # stub & ruby.exe within same folder.  Portable
       <<-TEXT
 @ECHO OFF
@@ -926,7 +931,7 @@ TEXT
     build_info_dir = File.join gem_home, "build_info"
 
     dir_mode = options[:dir_mode]
-    FileUtils.mkdir_p build_info_dir, :mode => dir_mode && 0o755
+    FileUtils.mkdir_p build_info_dir, mode: dir_mode && 0o755
 
     build_info_file = File.join build_info_dir, "#{spec.full_name}.info"
 
@@ -959,6 +964,19 @@ TEXT
 
   private
 
+  def user_install_dir
+    # never install to user home in --build-root mode
+    return unless @build_root.nil?
+
+    # Please note that @user_install might have three states:
+    # * `true`: `--user-install`
+    # * `false`: `--no-user-install` and
+    # * `nil`: option was not specified
+    if @user_install || (@user_install.nil? && Gem.default_user_install)
+      Gem.user_dir
+    end
+  end
+
   def build_args
     @build_args ||= begin
                       require_relative "command"
@@ -980,18 +998,17 @@ TEXT
 
   def bash_prolog_script
     if load_relative_enabled?
-      script = +<<~EOS
-        bindir="${0%/*}"
-      EOS
-
-      script << %Q(exec "$bindir/#{ruby_install_name}" "-x" "$0" "$@"\n)
-
       <<~EOS
         #!/bin/sh
         # -*- ruby -*-
         _=_\\
         =begin
-        #{script.chomp}
+        bindir="${0%/*}"
+        ruby="$bindir/#{ruby_install_name}"
+        if [ ! -f "$ruby" ]; then
+          ruby="#{ruby_install_name}"
+        fi
+        exec "$ruby" "-x" "$0" "$@"
         =end
       EOS
     else
